#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
options:
  action:
    description: What to do, audit, harden or rollback
    default: audit
  backupfile:
    description: Storage system backup configuration
    default: None
  path:
    description: Get the full path of the file/object of the fact
    default: null
  follow:
    description: Whether to follow symbolic links
    default: no
  owner:
    description: Numeric id representing the file owner
    type: int
  group:
    description: Numeric id representing the group of the owner
    type: int
  mode:
    description: Unix permissions of the file in octal
    type: octal
'''

import json
import os
import stat
import sys
import traceback
import grp
import pwd

from modules.module_base import ModuleBase


def get_user_name(uid):
    users = pwd.getpwall()
    for user in users:
        if user.pw_uid == uid:
            return user.pw_name
    return uid


def get_user_id(name):
    users = pwd.getpwall()
    for user in users:
        if user.pw_name == name:
            return user.pw_uid
    return None


def get_group_name(gid):
    groups = grp.getgrall()
    for group in groups:
        if group.gr_gid == gid:
            return group.gr_name
    return gid


def get_group_id(name):
    groups = grp.getgrall()
    for group in groups:
        if group.gr_name == name:
            return group.gr_gid
    return None


def mychmod(path, mode):
    oct_mode = int(mode, 8)
    exp_mode = 0
    if oct_mode & stat.S_IRUSR == stat.S_IRUSR:
        exp_mode = exp_mode | stat.S_IRUSR
    if oct_mode & stat.S_IWUSR == stat.S_IWUSR:
        exp_mode = exp_mode | stat.S_IWUSR
    if oct_mode & stat.S_IXUSR == stat.S_IXUSR:
        exp_mode = exp_mode | stat.S_IXUSR
    if oct_mode & stat.S_IRGRP == stat.S_IRGRP:
        exp_mode = exp_mode | stat.S_IRGRP
    if oct_mode & stat.S_IWGRP == stat.S_IWGRP:
        exp_mode = exp_mode | stat.S_IWGRP
    if oct_mode & stat.S_IXGRP == stat.S_IXGRP:
        exp_mode = exp_mode | stat.S_IXGRP
    if oct_mode & stat.S_IROTH == stat.S_IROTH:
        exp_mode = exp_mode | stat.S_IROTH
    if oct_mode & stat.S_IWOTH == stat.S_IWOTH:
        exp_mode = exp_mode | stat.S_IWOTH
    if oct_mode & stat.S_IXOTH == stat.S_IXOTH:
        exp_mode = exp_mode | stat.S_IXOTH
    if oct_mode & stat.S_ISUID == stat.S_ISUID:
        exp_mode = exp_mode | stat.S_ISUID
    if oct_mode & stat.S_ISGID == stat.S_ISGID:
        exp_mode = exp_mode | stat.S_ISGID

    os.chmod(path, exp_mode)


def run(arguments):
    module = ModuleBase(
        argument_spec=dict(
            action=dict(required=False, default='audit',
                        choices=['audit', 'harden', 'rollback']),
            backupfile=dict(required=False, default=None),
            path=dict(required=True, default=None),
            follow=dict(default=False, type='bool'),
            owner=dict(default=None),
            group=dict(default=None),
            mode=dict(default=None),
            owner_r=dict(default=None, type='bool'),
            owner_w=dict(default=None, type='bool'),
            owner_x=dict(default=None, type='bool'),
            group_r=dict(default=None, type='bool'),
            group_w=dict(default=None, type='bool'),
            group_x=dict(default=None, type='bool'),
            other_r=dict(default=None, type='bool'),
            other_w=dict(default=None, type='bool'),
            other_x=dict(default=None, type='bool'),
            suid=dict(default=None, type='bool'),
            sgid=dict(default=None, type='bool'),
            stick=dict(default=None, type='bool'),
        ))

    rc_flag, value = module.load_params(arguments)
    if rc_flag != 0:
        return module.fail_json(rc_flag=256, msg=value)

    # check which level need to process
    _action = module.params['action']
    _backupfile = module.params['backupfile']
    _follow = module.params['follow']
    _path = module.params['path']

    _ste_values = {'path': _path}
    _ste_values['follow'] = _follow
    _ste_values['mode'] = module.params['mode']
    _ste_values['owner'] = module.params['owner']
    _ste_values['group'] = module.params['group']
    _ste_values['owner_r'] = module.params['owner_r']
    _ste_values['owner_w'] = module.params['owner_w']
    _ste_values['owner_x'] = module.params['owner_x']
    _ste_values['group_r'] = module.params['group_r']
    _ste_values['group_w'] = module.params['group_w']
    _ste_values['group_x'] = module.params['group_x']
    _ste_values['other_r'] = module.params['other_r']
    _ste_values['other_w'] = module.params['other_w']
    _ste_values['other_x'] = module.params['other_x']
    _ste_values['suid'] = module.params['suid']
    _ste_values['sgid'] = module.params['sgid']
    _ste_values['stick'] = module.params['stick']

    # get current state
    # main stat data
    curr_state = None
    try:
        if _follow:
            curr_state = os.stat(_path)
        else:
            curr_state = os.lstat(_path)
    except OSError:
        err = sys.exc_info()[1]
        return module.fail_json(msg=err.strerror)

    st_mode = curr_state.st_mode
    item_values = {}
    item_values['path'] = module.params['path']
    item_values['mode'] = "%04o" % stat.S_IMODE(st_mode)
    item_values['st_mode'] = st_mode
    item_values['owner'] = get_user_name(curr_state.st_uid)
    item_values['group'] = get_group_name(curr_state.st_gid)
    item_values['uid'] = curr_state.st_uid
    item_values['gid'] = curr_state.st_gid
    item_values['owner_r'] = bool(st_mode & stat.S_IRUSR)
    item_values['owner_w'] = bool(st_mode & stat.S_IWUSR)
    item_values['owner_x'] = bool(st_mode & stat.S_IXUSR)
    item_values['group_r'] = bool(st_mode & stat.S_IRGRP)
    item_values['group_w'] = bool(st_mode & stat.S_IWGRP)
    item_values['group_x'] = bool(st_mode & stat.S_IXGRP)
    item_values['other_r'] = bool(st_mode & stat.S_IROTH)
    item_values['other_w'] = bool(st_mode & stat.S_IWOTH)
    item_values['other_x'] = bool(st_mode & stat.S_IXOTH)
    item_values['suid'] = bool(st_mode & stat.S_ISUID)
    item_values['sgid'] = bool(st_mode & stat.S_ISGID)
    item_values['stick'] = bool(st_mode & stat.S_ISVTX)

    if item_values['owner'] is None:
        return module.fail_json(
            msg=("Failed to get username of uid : %s" % curr_state.st_uid))

    if item_values['group'] is None:
        return module.fail_json(
            msg=("Failed to get groupname of gid : %s" % curr_state.st_gid))

    # audit and harden need to do audit
    _ret_actions = {'path': _path}
    _ret_actions = {'follow': _follow}
    if _action == 'audit' or _action == 'harden':
        audit_resutl = run_audit(_ret_actions, _ste_values, item_values)

        if _action == 'audit':
            return module.exit_json(result=audit_resutl, items=_ret_actions,
                                    msg="")
        else:
            if audit_resutl == True:
                # not need to do harden
                return module.exit_json(result=True, items=_ret_actions,
                                        msg="Not need to harden.")
            else:
                result, message = run_harden(module, _backupfile, _ret_actions,
                                            _ste_values, item_values)
                if result == False:
                    return module.fail_json(msg=message)
                else:
                    return module.exit_json(result=True, items=_ret_actions,
                                            msg=message)

    elif _action == 'rollback':
        result, message = run_rollback(module, _backupfile, _ret_actions,
                                      _ste_values, item_values)
        if result == False:
            return module.fail_json(msg=message)
        else:
            return module.exit_json(result=True, items=_ret_actions, msg=message)
    else:
        return module.fail_json(
            msg='Support action type is audit/harden/rollback.')


def run_audit(ret_actions, ste_values, item_values):
    audit_resutl = True

    # check owner
    if ste_values['owner'] is not None:
        ret_actions['owner'] = item_values['owner']
        # check is used !username
        if ste_values['owner'].startswith('!'):
            tmp_owner = ste_values['owner'].strip('!')
            if tmp_owner == item_values['owner']:
                audit_resutl = False
        else:
            if ste_values['owner'] != item_values['owner']:
                audit_resutl = False

    # check group
    if ste_values['group'] is not None:
        ret_actions['group'] = item_values['group']
        if ste_values['group'].startswith('!'):
            tmp_group = ste_values['group'].strip('!')
            if tmp_group == item_values['group']:
                audit_resutl = False
        else:
            if ste_values['group'] != item_values['group']:
                audit_resutl = False

    # check mode
    for key in ['mode', 'owner_r', 'owner_w', 'owner_x', 'group_r', 'group_w',
                'group_x', 'other_r', 'other_w',
                'other_x', 'suid', 'sgid', 'stick']:
        if ste_values[key] is not None:
            ret_actions[key] = item_values[key]
            if ste_values[key] != item_values[key]:
                audit_resutl = False

    ret_actions['mode'] = item_values['mode']

    return audit_resutl


def run_harden(module, backupfile, ret_actions, ste_values, item_values):
    needbackup = False

    try:
        # check need to backup
        if backupfile is None:
            return False, "You need to specify a backup file before harden."

        # if the file is not exist, neet to backup
        backupfile = os.path.realpath(backupfile)
        if os.path.exists(backupfile) is True:
            needbackup = False
        else:
            needbackup = True
            parentpath = os.path.dirname(backupfile)
            if not os.path.exists(parentpath):
                os.makedirs(parentpath)

        path = ste_values['path']
        follow = ste_values['follow']
        backup_json = {'path': path}
        backup_json['follow'] = follow
        backup_json['owner'] = item_values['owner']
        backup_json['group'] = item_values['group']
        backup_json['mode'] = item_values['mode']

        if needbackup is True:
            try:
                bkfp = open(backupfile, 'w')
                os.chmod(backupfile, int('600', 8))
                json.dump(backup_json, bkfp, indent=4)
                bkfp.close()
            except:
                return False, ('Failed to back configration to file %s.' % (
                    backupfile))

        if ste_values['owner'] is not None or ste_values['group'] is not None:
            if follow:
                os.chown(path, item_values['uid'], item_values['gid'])
            else:
                os.lchown(path, item_values['uid'], item_values['gid'])

        if ste_values['mode'] is not None:
            mychmod(path, ste_values['mode'])

        exp_mode = item_values['st_mode']
        if ste_values['owner_r'] is not None:
            if ste_values['owner_r']:
                exp_mode = exp_mode | stat.S_IRUSR
            else:
                exp_mode = exp_mode & (~stat.S_IRUSR)
        if ste_values['owner_w'] is not None:
            if ste_values['owner_w']:
                exp_mode = exp_mode | stat.S_IWUSR
            else:
                exp_mode = exp_mode & (~stat.S_IWUSR)
        if ste_values['owner_x'] is not None:
            if ste_values['owner_x']:
                exp_mode = exp_mode | stat.S_IXUSR
            else:
                exp_mode = exp_mode & (~stat.S_IXUSR)
        if ste_values['group_r'] is not None:
            if ste_values['group_r']:
                exp_mode = exp_mode | stat.S_IRGRP
            else:
                exp_mode = exp_mode & (~stat.S_IRGRP)
        if ste_values['group_w'] is not None:
            if ste_values['group_w']:
                exp_mode = exp_mode | stat.S_IWGRP
            else:
                exp_mode = exp_mode & (~stat.S_IWGRP)
        if ste_values['group_x'] is not None:
            if ste_values['group_x']:
                exp_mode = exp_mode | stat.S_IXGRP
            else:
                exp_mode = exp_mode & (~stat.S_IXGRP)
        if ste_values['other_r'] is not None:
            if ste_values['other_r']:
                exp_mode = exp_mode | stat.S_IROTH
            else:
                exp_mode = exp_mode & (~stat.S_IROTH)
        if ste_values['other_w'] is not None:
            if ste_values['other_w']:
                exp_mode = exp_mode | stat.S_IWOTH
            else:
                exp_mode = exp_mode & (~stat.S_IWOTH)
        if ste_values['other_x'] is not None:
            if ste_values['other_x']:
                exp_mode = exp_mode | stat.S_IXOTH
            else:
                exp_mode = exp_mode & (~stat.S_IXOTH)
        if ste_values['suid'] is not None:
            if ste_values['suid']:
                exp_mode = exp_mode | stat.S_ISUID
            else:
                exp_mode = exp_mode & (~stat.S_ISUID)
        if ste_values['sgid'] is not None:
            if ste_values['sgid']:
                exp_mode = exp_mode | stat.S_ISGID
            else:
                exp_mode = exp_mode & (~stat.S_ISGID)
        if ste_values['stick'] is not None:
            if ste_values['stick']:
                exp_mode = exp_mode | stat.S_ISVTX
            else:
                exp_mode = exp_mode & (~stat.S_ISVTX)

        if exp_mode != item_values['st_mode']:
            os.chmod(path, exp_mode)

        return True, ""
    except:
        return False, traceback.format_exc()


def run_rollback(module, backupfile, ret_actions, ste_values, item_values):
    if backupfile is None:
        return False, "You need to specify a backup file when rollback."

    backupfile = os.path.realpath(backupfile)
    if os.path.exists(backupfile) is False:
        return False, "You need to specify a backup file when rollback."

    backup_json = None

    with open(backupfile, 'r') as bkfp:
        backup_json = json.load(bkfp)

    try:
        path = backup_json['path']
        follow = backup_json['follow']

        if path is None or os.path.exists(path) is False:
            return False, ("File %s is not exist." % path)

        needcount = 0
        if 'owner' in backup_json:
            ret_actions['owner'] = backup_json['owner']
            needcount = needcount + 1

        if 'group' in backup_json:
            ret_actions['group'] = backup_json['group']
            needcount = needcount + 1

        if needcount == 1:
            return False, "The owner or group information is lacking in the " \
                          "backup file."

        uid = item_values['uid']
        gid = item_values['gid']

        if needcount == 2:
            if backup_json['owner'] != item_values['owner'] or backup_json[
                'group'] != item_values['group']:
                uid = get_user_id(backup_json['owner'])
                gid = get_group_id(backup_json['group'])

                if uid is None:
                    return False, ('The user of uid %s is not exist.' % uid)
                if gid is None:
                    return False, ('The group of gid %s is not exist.' % gid)

                if follow:
                    os.chown(ste_values['path'], uid, gid)
                else:
                    os.lchown(ste_values['path'], uid, gid)

        if backup_json['mode'] is not None:
            ret_actions['mode'] = backup_json['mode']
            mychmod(backup_json['path'], backup_json['mode'])

        return True, ""
    except:
        return False, traceback.format_exc()
